/*
 * module-itip-formatter-web-extension.h
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the program; if not, see <http://www.gnu.org/licenses/>
 *
 */

#ifndef MODULE_ITIP_FORMATTER_WEB_EXTENSION_H
#define MODULE_ITIP_FORMATTER_WEB_EXTENSION_H

#define MODULE_ITIP_FORMATTER_WEB_EXTENSION_SERVICE_NAME "org.gnome.Evolution.Module.ItipFormatter.WebExtension"
#define MODULE_ITIP_FORMATTER_WEB_EXTENSION_OBJECT_PATH  "/org/gnome/Evolution/Module/ItipFormatter/WebExtension"
#define MODULE_ITIP_FORMATTER_WEB_EXTENSION_INTERFACE    "org.gnome.Evolution.Module.ItipFormatter.WebExtension"

#endif /* MODULE_ITIP_FORMATTER_WEB_EXTENSION_H */
